import ftplib
from ftplib import FTP
import sys
import os
import pandas as pd
from datetime import datetime, timedelta
from tqdm.notebook import tqdm
from zkyhaxpy.io_tools import create_folders


def ftp_connection(ftp_address, user, password):
    '''
    Create a FTP connection
	'''
    
    ftp_conn = FTP(ftp_address)  # connect to host, default port
    #ftp.connect(port=21)       
    ftp_conn.login(user=user, passwd=password)# user anonymous, passwd anonymous@


    return ftp_conn



def ftp_walk(ftp_conn, dir, df_ftp_path):
    '''
    List all files (recursively) and save into df dataframe. df is required to define as global object.
    '''
    dirs = []
    nondirs = []
    for item in ftp_conn.mlsd(dir):
        if item[1]['type'] == 'dir':
            dirs.append(item[0])            
        else:            
            if item[0] not in ('.', '..'):
                nondirs.append(item[0])
                id = len(df_ftp_path)
                df_ftp_path.at[id, 'dir'] = dir
                df_ftp_path.at[id, 'filename'] = item[0]
                str_mod_date = str(item[1]['modify'])[:14]          
                #df.at[id, 'mod_date'] = datetime.strptime(str_mod_date, '%Y%m%d')
                df_ftp_path.at[id, 'mod_date'] = datetime.strptime(str_mod_date, '%Y%m%d%H%M%S') + timedelta(hours=7)
    if nondirs:        
        print('{} : {} files.'.format(dir,len(nondirs)))
        #print('\n'.join(sorted(nondirs)))
        
    else:
        # print(dir, 'is empty')
        pass
    for subdir in sorted(dirs):
        ftp_walk(ftp_conn, '{}/{}'.format(dir, subdir), df_ftp_path)


def ftp_download(ftp_conn, ftp_path, out_path, skip_exist=True):
    '''
    Download file from specified ftp path.
	'''
    if (os.path.exists(out_path)==True) & (skip_exist==True):
        print(f'{out_path} already exists.')
        return

    create_folders(out_path)
    with open(out_path, 'wb') as fp:        
        ftp_conn.retrbinary(f'RETR {ftp_path}', fp.write)
        print(f'{ftp_path} has been downloaded.')
    




import pandas as pd
import numpy as np
from IPython.display import display, HTML, display_html
from tqdm.notebook import tqdm

def auto_adjust():
    '''
    Set column width = 100
    Max displayed rows = 100
    Max displayed columns = 100
	'''
    set_colwidth(100)
    pd.options.display.max_rows = 100
    pd.options.display.max_columns = 100



def set_max_rows(max_rows=100):
    '''
    Set max display rows
    
    Return : None
    '''

    pd.options.display.max_rows = max_rows


def set_max_columns(max_columns=100):
    '''
    Set max display columns
    
    Return : None
    '''

    pd.options.display.max_columns = max_columns



def display_html(df):
    '''
    display a dataframe as html table
    '''
    display(HTML(df.to_html()))


def inc_colwidth(inc_colwidth=20,target_colwidth=None):
    '''
    Increase column width of pandas dataframe display

    Return : None

    '''
    if target_colwidth == None:
        curr_max_colwidth = pd.get_option("display.max_colwidth") 
        new_max_colwidth = curr_max_colwidth + inc_colwidth
        pd.set_option('max_colwidth', new_max_colwidth)
    else:
        pd.set_option('max_colwidth', target_colwidth)

    print(f'Current max column width = {pd.get_option("display.max_colwidth")}')


def dec_colwidth(dec_colwidth=20,target_colwidth=None):
    '''
    Decrease column width of pandas dataframe display
    
    Return : None
    '''
    if target_colwidth == None:
        curr_max_colwidth = pd.get_option("display.max_colwidth") 
        new_max_colwidth = curr_max_colwidth - dec_colwidth
        pd.set_option('max_colwidth', new_max_colwidth)
    else:
        pd.set_option('max_colwidth', target_colwidth)   

    print(f'Current max column width = {pd.get_option("display.max_colwidth")}')


def set_colwidth(target_colwidth=100):
    '''
    Decrease column width of pandas dataframe display
    
    Return : None
    '''

    pd.set_option('max_colwidth', target_colwidth)


def get_curr_colwidth():
    '''
    Decrease column width of pandas dataframe display
    
    Return : None
    '''

    print(f'Current max column width = {pd.get_option("display.max_colwidth")}')
    
    



        
def read_parquets(list_file_path, columns='all'):
    '''

    Read multiple parquet files of the same template into a single pandas dataframe.

	'''
    
    list_df = []
    for file_path in tqdm(list_file_path, 'reading parquets...'):
        if columns=='all':
            list_df.append(pd.read_parquet(file_path))
        else:
            list_df.append(pd.read_parquet(file_path, columns=columns))
        
    df = pd.concat(list_df)
    
    return df


def convert_dtypes(in_df, in_dict_dtypes, default_dtype=None):
    '''
    Convert dtypes of a dataframe according to given dict of column names and dtypes.
    '''
    in_df = in_df.copy()
    for col_nm in in_df.columns:
        if col_nm in in_dict_dtypes.keys():
            if in_df[col_nm].dtype != in_dict_dtypes[col_nm]:
                in_df[col_nm] = in_df[col_nm].astype(in_dict_dtypes[col_nm])
        elif default_dtype:
            if in_df[col_nm].dtype != default_dtype:
                in_df[col_nm] = in_df[col_nm].astype(default_dtype)
            
    return in_df




def optimize_dtypes(df, excluded_cols=None, only_int=True, allow_unsigned=False):
    
    '''
    Optimize data type of each column to minimum size.
    '''
    df = df.copy()
    
    if excluded_cols:
        assert(type(excluded_cols) == list)
        list_cols = [col for col in df.columns if col not in excluded_cols]
        
    else:
        list_cols = list(df.columns)

 

    
    if (only_int==True) :
        list_cols = [col for col in list_cols if 'int' in str(df[col].dtype)]
        
        
    for col in list_cols:
        col_dtype_ori_str = str(df[col].dtype)        
        col_max_val = df[col].max()
        col_min_val = df[col].min()

 

        if 'int' in col_dtype_ori_str:
            if (col_min_val >= 0) & (allow_unsigned==True):
                if col_max_val < 2**8:
                    col_dtype_new = np.uint8
                elif col_max_val < 2**16:
                    col_dtype_new = np.uint16
                elif col_max_val < 2**32:
                    col_dtype_new = np.uint32
                else:
                    col_dtype_new = np.uint64                    
                    
            else:
                if (col_max_val < 2**7) & (col_min_val >= -2**7):
                    col_dtype_new = np.int8
                elif (col_max_val < 2**15) & (col_min_val >= -2**15):
                    col_dtype_new = np.int16
                elif (col_max_val < 2**31) & (col_min_val >= -2**31):
                    col_dtype_new = np.int32
                else:
                    col_dtype_new = np.int64
                    
            
            assert(col_min_val == col_dtype_new(col_min_val))
            assert(col_max_val == col_dtype_new(col_max_val))            
            col_dtype_new_str = str(col_dtype_new).split("'")[1].split('.')[1]
            if col_dtype_ori_str != col_dtype_new_str:
                df[col] = df[col].astype(col_dtype_new)
                print(f'Column "{col}": {col_dtype_ori_str} -> {col_dtype_new_str}')
        else:
            pass
    
    return df



import json
import os


def read_dict_json(json_path, print_result=False):   
    '''
    Read a json file into a dictionary.
    '''
    with open(json_path, 'r') as f:
        out_dict = json.load(f)
    if print_result:
        print(f'{json_path} has been loaded.')
    return out_dict


def write_dict_json(in_dict, json_path, force_overwrite=False):   
    '''
    Write a dictionary into a json file.
    '''    

    #Check if the output file already exists and force_overwrite is false
    if (os.path.exists(json_path)==True) & (force_overwrite==False):
        str_input = input(f'"{json_path}" already exists.\nDo you want to overwrite it? (YES/NO)')
        if (str_input != 'YES'):
            #Do nothing
            return 
            
    with open(json_path, 'w') as f:
        f.write(json.dumps(in_dict, indent=2))
        print(f'{json_path} has been saved.')
    return 


import datetime

class timer:
  
    def __init__(self, name):
        self.name = name
        self.start_time = None
        self.end_time = None
        self.lap_times = []
        self.elapsed_time = None

    def start(self):
        self.start_time = datetime.datetime.now()
        print(f'Start at: {self.start_time}')

    def end(self):
        self.end_time = datetime.datetime.now()        
        print(f'End at: {self.end_time}')
        self.elapsed_time = self.end_time - self.start_time
        print()
        self.show()

    def lap(self):
        curr_lap_time = datetime.datetime.now()
        if len(self.lap_times) > 0:
            prev_lap_time = self.lap_times[-1]
        else:
            prev_lap_time = self.start_time
        latest_lap_elapsed_time = curr_lap_time - prev_lap_time
        self.lap_times.append(curr_lap_time)
        
        print(f'Lap {len(self.lap_times)} time at: {curr_lap_time}')
        print(f'Lap {len(self.lap_times)} time: {latest_lap_elapsed_time}')

    def show(self):
        print('###############################')
        print(f'Timer: {self.name}')          
        if len(self.lap_times) > 0:
            for i, lap_time in enumerate(self.lap_times):
                
                if i == 0:
                    curr_lap_elapsed_time = lap_time - self.start_time
                else:
                    curr_lap_elapsed_time = lap_time - prev_lap_time
                prev_lap_time = lap_time                
                [print(f'Lap {i+1} elapsed time: {curr_lap_elapsed_time}')]
            print(f'Lap {i+2} elapsed time: {self.end_time - prev_lap_time}')
            
        print(f'Total elapsed time: {self.elapsed_time}')
        print('###############################')

timer_1 = timer('Test')



# python3
import hashlib


def encrypt_string(input_string, salt):
    # Assume cleaned input string

    # Append salt to the end of input string
    input_string = input_string + salt

    # encrypt using utf-8 encoding
    encrypted_string = hashlib.sha256(input_string.encode('utf8')).hexdigest()

    return encrypted_string





import os
from google.colab import auth
from google.colab import drive
from zkyhaxpy.io_tools import get_list_files_re, filepaths_to_df
import pandas as pd
import re


def check_dup_files(folder, file_nm_prefix=None, file_nm_extension=None, remove=False):
    if file_nm_prefix==None:
        file_nm_prefix = '.*'
    if file_nm_extension==None:
        file_nm_extension = '.*'
        
    list_file_paths = get_list_files_re(folder, file_nm_prefix + '.*\([0-9]{1,}\)\.' + file_nm_extension)    
    
    if remove==True:
        print(f'Total of {len(list_file_paths)} duplicated files will be removed.')
        for filepath in list_file_paths:
            os.remove(filepath)
            print(f'{filepath} is removed.')
    else:   
        return list_file_paths
        
    
def mount_drive():
    drive.mount('/content/drive', force_remount=True)
    
def authen_gcp():
    auth.authenticate_user()




def get_list_files_re_gcs(bucket_path_prefix, storage_client, filename_re=None, folder_re=None, return_as_df=False):
    '''
    Get a list of files in Google Cloud Storage

    Parameters
    -------------------------------------
    bucket_path_prefix (str): 
        prefix of files to be listed (must start with bucket name)
        For example, if bucket name = 'test-bucket', and a file is saved at '/test-bucket/path_prefix/test1.file'
        To list all files in '/test-bucket/path_prefix/', path_prefix shall be '/test-bucket/path_prefix/'

    storage_client (google.cloud.storage.client.Client):
        Storage client that can be used to access the bucket

    filename_re (str): 
        regular expression to search for filename

    folder_re (str): 
        regular expression to search for folder


    Output
    -------------------------------------
    return : a list of tuple(filepaths
    '''

    bucket_name = bucket_path_prefix.split('/')[1]
    path_prefix = bucket_path_prefix[len(bucket_name)+2:]
    bucket = storage_client.bucket(bucket_name)

    all_blobs = list(bucket.list_blobs(prefix=path_prefix))

    list_files = []
    if len(all_blobs) == 0:
        return list_files


    df_files = pd.DataFrame(all_blobs, columns=['blob'])
    df_files['path'] = df_files['blob'].astype(str).str.split(',', expand=True).loc[:, 1].str.strip()  

    if filename_re == None:
        filename_re = '.*'
    if folder_re == None:
        folder_re = '.*'
    for filepath in list(df_files['path']):   
        file_nm = os.path.basename(filepath)
        folder = os.path.dirname(filepath)
        if ((re.search(filename_re, file_nm) != None) & (re.search(folder_re, folder) != None)):
            list_files.append(os.path.join('/'+bucket_name, folder, file_nm))

        
    if return_as_df==False:
        return list_files
    else:
        return filepaths_to_df(list_files)    
        
    


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zkyhaxpy", # Replace with your own username
    version="0.2.9.3.7",
    author="Surasak Choedpasuporn",
    author_email="surasak.cho@gmail.com",
    description="A swiss-knife Data Science package for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/surasakcho/zkyhaxpy",
    packages=setuptools.find_packages(),    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




import numpy as np
from skimage import exposure

def image_from_separated_rgb(image_r, image_g, image_b, clip_min_pct=2, clip_max_pct=98, equalize_hist=True):
    '''
    return array(m, n, 3) of RGB image (0-255)
    '''
    image_b = np.clip(image_b, np.percentile(image_b, clip_min_pct), np.percentile(image_b, clip_max_pct))
    image_g = np.clip(image_g, np.percentile(image_g, clip_min_pct), np.percentile(image_g, clip_max_pct))
    image_r = np.clip(image_r, np.percentile(image_r, clip_min_pct), np.percentile(image_r, clip_max_pct))

    if equalize_hist == True:
        image_b = exposure.equalize_hist(image_b)
        image_g = exposure.equalize_hist(image_g)
        image_r = exposure.equalize_hist(image_r)

    image_b = (image_b - image_b.min()) / (image_b.max() - image_b.min())
    image_g = (image_g - image_g.min()) / (image_g.max() - image_g.min())
    image_r = (image_r - image_r.min()) / (image_r.max() - image_r.min())

    image_b = image_b * 255
    image_g = image_g * 255
    image_r = image_r * 255

    # image = make_lupton_rgb(image_r, image_g, image_b, stretch=5, Q=5)
    image = np.array(list(zip(image_r.flatten(), image_g.flatten(), image_b.flatten())), dtype=np.int16).reshape(*image_r.shape, 3)
    return image


import numpy as np

def case_when(*args, default=np.nan, dtype=float): 
    '''

    SQL case when condition like for numpy array.

    Parameters
    ----------
    args: a numpy array
        A numpy array of lenght 2 * number of cases. With the first element of a pair is array of boolen and the second element of a pair is value if the condition is matched.
        
    default: int or float
        A value if there is none condition is matched.
    
    dtype: dtype
        dtype of the output array
        

    Examples
    --------
    arr = np.array([0, 2, 5, 3, 4, 10, 4])
    case_when(
        (np.isin(arr, [5, 10]), 5),
        (np.isin(arr, [2, 4, 6, 8, 10]), 2),
        (arr == 0, 0),
        default=-999   
    )

    Returns
    -------
    Numpy array of the same shape as inputted conditions with element assigned according to each case when condition.

	'''
    
    out_shape = args[0][0].shape    
    arr_out = np.full(out_shape, np.nan)
    arr_non_default_mask = np.full(out_shape, False) 

    for i in range(len(args)):
        assert(out_shape == args[i][0].shape)
        arr_out = np.where((np.isnan(arr_out) & (args[i][0])) , args[i][1], arr_out)                
        arr_non_default_mask = np.where(args[i][0], True, arr_non_default_mask)

    arr_out = np.where(arr_non_default_mask, arr_out, default)
    return arr_out.astype(dtype)


def fillna(in_arr, in_fillvalue):
    '''

    Fill nan value in numpy array with fill in value(s).

    Parameters
    ----------
    in_arr: a numpy array
    in_fillvalue : a scalar or numpy array with the same shape of in_arr
        
        Returns
    -------
    out_arr: a numpy array that already filled

	'''
    out_arr = np.where(
        np.isnan(in_arr),
        in_fillvalue,
        in_arr)
    
    return out_arr
    
    
    




def get_last_n_digit(in_arr, nbr_digits=1):    
    '''
    Get N last digit(s) of the given array.
    '''
    
    assert(type(nbr_digits)==int)
    assert(nbr_digits >= 1)
    assert(nbr_digits <= 4)
    divider = 10 ** nbr_digits
    out_arr = np.remainder(in_arr, divider).astype(int)
    
    return out_arr    


def get_mod(in_arr, mod):    
    '''
    Get mod for given array
    '''
    out_arr = np.remainder(in_arr, mod).astype(int)
    
    return out_arr